﻿using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Linq;
using System.Text;
using System.Threading;

namespace PlatformSDKSample
{
    using TEventInfo = PlatformSDK.TEventInfo;
    using EEventType = PlatformSDK.EEventType;
    using EDeviceType = PlatformSDK.EDeviceType;
    using TDeviceInfo = PlatformSDK.TDeviceInfo;
    using EConnectionStatus = PlatformSDK.EConnectionStatus;
    using ERecordingStatus = PlatformSDK.ERecordingStatus;
    using TDeviceStatusInfo = PlatformSDK.TDeviceStatusInfo;

    public partial class MainUI : Form
    {
        string GetDeviceStatusDesc(EConnectionStatus status)
        {
            switch (status)
            {
                case EConnectionStatus.connected:
                    return "connected";
                case EConnectionStatus.disconnected:
                    return "disconnected";
                case EConnectionStatus.unchanged:
                    return "unchanged";
            }
            return "unknow";
        }

        string GetEventDesc(TEventInfo info)
        {
            switch (info.eventType)
            {
                case EEventType.DO: 
                    return "DO-" + info.index + (info.trigger == 1 ? " trigger" : " normal");
                case EEventType.DI:
                    return "DI-" + info.index + (info.trigger == 1 ? " trigger" : " normal");
                case EEventType.motion:
                    return "Motion-" + info.index + " percent: " + info.args;
                case EEventType.tampering:
                    return "Tampering";
                case EEventType.PIR:
                    return "PIR-" + info.index;
                case EEventType.videoLoss:
                    return (info.trigger == 1 ? "Video loss" : "Video restore");
                case EEventType.fanStatus:
                    return "Fan-" + info.index + (info.trigger == 1 ? " error" : " fine");
                case EEventType.storageConn:
                    return "StorageConn-Volume " + info.index + (info.trigger == 0 ? "available" : (info.trigger == 1 ? "Unavailable" : "Degrade"));
                case EEventType.storageCap:
                    return "StorageCap " + (info.trigger == 1 ? "Exhausted" : "Healthy");
                case EEventType.storageVoid:
                    return "StorageVoid " + (info.trigger == 1 ? "Exhausted" : "Healthy");
            }
            return "Unknown";
        }

        void UpdateEventWindow(TEventInfo eventInfo)
        {
            ListViewItem lvi = new ListViewItem();

            lvi.Text = GetDeviceDisplayName(eventInfo.deviceID);
            var init = eventInfo.init == 1 ? "(Init)" : "";
            lvi.SubItems.Add(init + GetEventDesc(eventInfo));
            lvi.SubItems.Add(eventInfo.time);

            eventWindow.Items.Add(lvi);
        }

        TreeNode FindDeviceTreeNode(string deviceID)
        {
            if (treeViewDevices.Nodes.Count == 0)
            {
                return null;
            }

            foreach (TreeNode deviceType in treeViewDevices.Nodes[0].Nodes)
            {
                foreach (TreeNode device in deviceType.Nodes)
                {
                    if (device.Tag == null) continue;
                    if (((TDeviceInfo)device.Tag).deviceID == deviceID) return device;
                }
            }

            return null;
        }

        TreeNode GetDeviceTreeNode(TEventInfo eventInfo)
        {
            TreeNode device = FindDeviceTreeNode(eventInfo.deviceID);
            if (device == null)
            {
                Console.WriteLine(System.Reflection.MethodBase.GetCurrentMethod().Name
                    + " fail: "
                    + eventInfo.deviceID + ", "
                    + GetEventDesc(eventInfo));
            }
            return device;
        }

        string ComposeDiDoNodeName(TEventInfo eventInfo)
        {
            return (eventInfo.trigger == 1 ? "● " : "○ ")
                + (eventInfo.eventType == EEventType.DI ? "DI-" : "DO-")
                + eventInfo.index;
        }

        // The ID is used for speed-up lookup for DI/DO node and get its status
        // Set DI1 ~ DI99 to 10~990, set DO1 ~ DO99 to 1010~1990
        // the ID of DI = index * 10 + trigger
        // the ID of DO = index * 10 + trigger + 1000
        int GetDiDoNodeID(TEventInfo eventInfo)
        {
            int ID = eventInfo.index * 10 + eventInfo.trigger;
            if (eventInfo.eventType == EEventType.DO) ID += 1000;
            return ID;
        }

        bool IsDONode(int DiDoNodeID)
        {
            return 1010 <= DiDoNodeID && DiDoNodeID <= 1991;
        }

        byte GetTriggerStatusFromDiDoNodeID(int ID)
        {
            return (byte)(ID % 10);
        }

        uint GetIndexFromDiDoNodeID(int ID)
        {
            uint index = (uint)(ID > 1000 ? ID - 1000 : ID);
            return index / 10;
        }

        void AddDiDoNodeToDevice(ref TreeNode device, TEventInfo eventInfo)
        {
            TreeNode node = new TreeNode(ComposeDiDoNodeName(eventInfo));
            node.Tag = GetDiDoNodeID(eventInfo);
            node.Name = DIDO_NODE;
            device.Nodes.Add(node);
        }

        bool IsValidDiDoIndex(TDeviceInfo deviceInfo, TEventInfo eventInfo)
        {
            if (eventInfo.deviceType == EDeviceType.camera) return true;

            switch (eventInfo.eventType)
            {
                case EEventType.DI: 
                    return eventInfo.index <= deviceInfo.DINum;
                case EEventType.DO: 
                    return eventInfo.index <= deviceInfo.DONum;
            }
            return false;
        }

        void AddDiDoNode(ref TreeNode device)
        {
            if (device.Tag == null) return;

            TDeviceInfo deviceInfo = (TDeviceInfo)device.Tag;
            string devieID = deviceInfo.deviceID;

            lock (didoNodeListLock)
            {
                foreach (TEventInfo eventInfo in didoNodeList)
                {
                    if (eventInfo.deviceID != devieID) continue;
                    if (!IsValidDiDoIndex(deviceInfo, eventInfo)) continue;
                    
                    AddDiDoNodeToDevice(ref device, eventInfo);
                    return;                    
                }
            }
        }

        void AddDiDoNode(TEventInfo eventInfo)
        {
            TreeNode device = GetDeviceTreeNode(eventInfo);
            if (device == null || device.Tag == null) return;

            TDeviceInfo deviceInfo = (TDeviceInfo)device.Tag;
            if (!IsValidDiDoIndex(deviceInfo, eventInfo)) return;

            AddDiDoNodeToDevice(ref device, eventInfo);

            lock (didoNodeListLock)
            {
                didoNodeList.Add(eventInfo);
            }
        }
        
        TreeNode FindDiDoNode(TEventInfo eventInfo)
        {
            TreeNode device = GetDeviceTreeNode(eventInfo);
            if (device == null) return null;

            int ID = GetDiDoNodeID(eventInfo) / 10;
            return device.Nodes.OfType<TreeNode>().
                FirstOrDefault(node => ((int)node.Tag) / 10 == ID);            
        }
        
        void UpdateDiDoNode(TEventInfo eventInfo)
        {
            TreeNode didoNode = FindDiDoNode(eventInfo);
            if (didoNode == null)
            {
                Console.WriteLine(System.Reflection.MethodBase.GetCurrentMethod().Name
                    + " fail: "
                    + eventInfo.deviceID + ", "
                    + GetEventDesc(eventInfo));
                AddDiDoNode(eventInfo);
                return;
            }

            didoNode.Text = ComposeDiDoNodeName(eventInfo);
            didoNode.Name = DIDO_NODE;
            didoNode.Tag = GetDiDoNodeID(eventInfo);
        }

        void UpdateDeviceStatusOnDeviceTree(TDeviceStatusInfo statusInfo)
        {
            TreeNode device = FindDeviceTreeNode(statusInfo.deviceID);
            if (device == null)
            {
                Console.WriteLine(System.Reflection.MethodBase.GetCurrentMethod().Name
                    + " fail: "
                    + statusInfo.deviceID + ", "
                    + GetDeviceStatusDesc(statusInfo.connectionStatus));
                return;
            }

            string treeNodeName = GetTreeNodeName((TDeviceInfo)(device.Tag));

            var preState = new string[] { "", "" };
            preState = device.ToolTipText == "" ? preState : device.ToolTipText.Split('/');

            var connectionStatus = preState[0];
            if (statusInfo.connectionStatus == EConnectionStatus.disconnected)
            {
                connectionStatus = "☒ ";
            }
            else if (statusInfo.connectionStatus == EConnectionStatus.connected)
            {
                connectionStatus = "";
            }

            var recordingStatus = preState[1];
            if (statusInfo.recordingStatus == ERecordingStatus.stop)
            {
                recordingStatus = "(stop)";
            }
            else if (statusInfo.recordingStatus == ERecordingStatus.start)
            {
                recordingStatus = "(start)";
            }

            device.ToolTipText = connectionStatus + "/" +recordingStatus ;
            device.Text = connectionStatus + treeNodeName + recordingStatus;
        }

        void UpdateDeviceTree(TEventInfo eventInfo)
        {
            if (eventInfo.eventType != EEventType.DI &&
                eventInfo.eventType != EEventType.DO) return;

            if (bReloginEventServer == 1)
            {
                UpdateDiDoNode(eventInfo);                
            }
            else
            {
                if (eventInfo.init == 1)
                {
                    AddDiDoNode(eventInfo);
                }
                else
                {
                    UpdateDiDoNode(eventInfo);
                }
            }
        }

        void UpdateEventWindowAndDeviceTree()
        {
            int count = 0;
            lock (eventListLock)
            {
                count = eventList.Count;
            }

            eventWindow.BeginUpdate();

            for (int i = 0; i < count; ++i)
            {
                UpdateEventWindow(eventList[i]);
                UpdateDeviceTree(eventList[i]);
            }

            eventWindow.EndUpdate();

            lock (eventListLock)
            {
                eventList.Clear();
            }            
        }

        void UpdateDeviceStatus()
        {
            int count = 0;
            lock (deviceStatusListLock)
            {
                count = deviceStatusList.Count;
            }

            for (int i = 0; i < count; ++i)
            {
                UpdateDeviceStatusOnDeviceTree(deviceStatusList[i]);
            }

            lock (deviceStatusListLock)
            {
                deviceStatusList.Clear();
            } 
        }
    }
}